/*
 *
 * Copyright (C) 2008 SpinetiX S.A.
 *
 * ----------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------------
 *
 */

/* Kernel log dump facility */

#include <common.h>
#include <rtc.h>
#include <malloc.h>
#include "klog-dump.h"

#ifdef CONFIG_KLOG_DUMP

#define KLOG_MAGIC 0x474F4C4B

#define KLOG_SECTOR_FULL	0x2a4f4c4b
#define KLOG_SECTOR_NONFULL	0x2e4f4c4b

/* #define DEBUG */

#ifdef DEBUG
#define DPRINTF(fmt, args...) printf(fmt,## args)
#else
#define DPRINTF(fmt, args...)
#endif

extern flash_info_t flash_info[];

static inline int is_addr_valid(unsigned long a)
{
     return (a >= KLOG_DATA_MIN_ADDR && a < KLOG_DATA_MAX_ADDR );
}

static inline ulong sector_size(flash_info_t *info, int sector)
{
	return (sector != (info->sector_count - 1)) ?
		(info->start[sector + 1] - info->start[sector]) :
		(info->start[0] + info->size - info->start[sector]);
}

static inline ulong trailing_free(flash_info_t *info, int sector)
{
	int n = 0;
	ulong start = info->start[sector];
	ulong end = start + sector_size(info, sector);
	
	while (--end >= start) {
		if ( *(unsigned char*) end != 0xFF )
			return n;
		n++;
	}
	return n;
}

/*
 * Reserves space for dumping the log in flash, making sure
 * that the dump can be written without erasing (recycles a sector
 * if necessary).
 * Returns the address where the entry is to be written, or zero
 * if none could be found.
 */
static ulong reserve_log_entry(ulong size)
{
	flash_info_t *info = &flash_info[0];
	int s_first, s_last;
	int s_target;
	ulong magic;

	if (info == NULL) {
		fprintf(stderr, "ERROR: klog dump region not in flash!\n");
		return 0;
	}

	s_first = ( FLASH_ADDR_KLOG - info->start[0] ) /
		CFG_FLASH_SECT_SZ;
	s_last = s_first + FLASH_SIZE_KLOG / CFG_FLASH_SECT_SZ - 1;

	for (s_target = s_first; s_target <= s_last; s_target++) {
		magic = *(ulong*)info->start[s_target];
		if ( magic != KLOG_SECTOR_FULL )
			break;
	}
	if ( s_target <= s_last ) {
		magic = *(ulong*)info->start[s_target];
		if ( magic == KLOG_SECTOR_NONFULL ) {
			ulong nfree = trailing_free(info, s_target);
			DPRINTF("found sector %i non-full with %lu free bytes\n",
				s_target, nfree);
			if (size <= nfree)
				return info->start[s_target] + sector_size(info, s_target) - nfree;
			DPRINTF("log entry of %lu bytes does not fit, marking as full\n", size);
			magic = KLOG_SECTOR_FULL;
			if ( write_buff(info, (uchar*)&magic, info->start[s_target], 4) )
				fprintf(stderr, "WARN: failed marking klog dump region as full\n");
			s_target = ( s_target != s_last ? s_target+1 : s_first );
		} else {
			DPRINTF("found sector %i not initialized\n", s_target);
		}
	} else {
		DPRINTF("found all sectors full\n");
		s_target = s_first;
	}

	DPRINTF("recycling sector %i\n", s_target);
	if ( trailing_free(info, s_target) != sector_size(info, s_target) ) {
		if ( flash_erase(info, s_target, s_target) ) {
			fprintf(stderr, "ERROR: failed klog dump region recycle\n");
			return 0;
		}
	}
	DPRINTF("marking sector %i non-full\n", s_target);
	magic = KLOG_SECTOR_NONFULL;
	if ( write_buff(info, (uchar*)&magic, info->start[s_target], 4) ) {
		fprintf(stderr, "ERROR: failed klog dump region recycle\n");
		return 0;
	}
	return info->start[s_target] + 4;

	return 0;
}

static void save_log(unsigned long log_end,
		     unsigned long logged_chars,
		     const char* log_buf,
		     unsigned long log_buf_len)
{
	const unsigned long idx_mask = log_buf_len - 1;
	unsigned long idx;
	struct rtc_time tm;
	char *out_buf;
	char *out_buf_end;
	ulong save_addr;

	if (logged_chars > KLOG_MAX_SIZE)
		logged_chars = KLOG_MAX_SIZE;

	out_buf = malloc(logged_chars + 64);
	if ( ! out_buf ) {
		fprintf(stderr, "ERROR: out of memory to dump klog\n");
		return;
	}

	printf("INFO: dumping %lu bytes of kernel logs to flash\n", logged_chars);

	out_buf_end = out_buf;
	out_buf_end += sprintf(out_buf, "\n- START %lu @ ", logged_chars);
	if (rtc_get(&tm) < 0)
		out_buf_end += sprintf(out_buf_end, "<unknown>\n");
	else
		out_buf_end += sprintf(out_buf_end,
				       "%4d-%02d-%02d %02d:%02d:%02dZ\n",
				       tm.tm_year, tm.tm_mon, tm.tm_mday,
				       tm.tm_hour, tm.tm_min, tm.tm_sec);
	for (idx = log_end - logged_chars; idx < log_end; idx++)
		*(out_buf_end++) = log_buf[idx & idx_mask];
	*(out_buf_end++) = '\n';
	out_buf_end += sprintf(out_buf_end, "- END\n");

#ifdef DEBUG
	puts(out_buf);
#endif

	save_addr = reserve_log_entry(out_buf_end - out_buf);
	if (save_addr) {
		flash_info_t *info = &flash_info[0];
		if ( write_buff(info, out_buf, save_addr, out_buf_end - out_buf) )
			fprintf(stderr, "ERROR: failed klog dump\n");
	}

	free(out_buf);
}


void klog_dump(void)
{
	volatile ulong *desc =
		(volatile ulong *)( KLOG_DUMP_ADDR );
	unsigned long log_end;
	unsigned long logged_chars;
	const char* log_buf;
	unsigned long log_buf_len;
	int ignore;

	if (desc[0] != KLOG_MAGIC) {
		DPRINTF("klog feature not initialized\n");
		return; /* feature not initialized */
	}
	if (desc[1] == 0 && desc[2] == 0 && desc[3] == 0 && desc[4] == 0) {
		DPRINTF("no klog data, ignoring\n");
		return; /* nothing present */
	}

	DPRINTF("klog raw data: 0x%lx 0x%lx 0x%lx 0x%lx\n",
		desc[1], desc[2], desc[3], desc[4]);

	/* Sanity check RAM addresses before dereferencing */
	if ( ! is_addr_valid(desc[1]) ||
	     ! is_addr_valid(desc[2]) ||
	     ! is_addr_valid(desc[3]) )
	{
		fprintf(stderr,
			"WARN: klog addresses out of range, ignoring\n");
		return;
	}
	log_end = *(const unsigned long*) desc[1];
	logged_chars = *(const unsigned long*) desc[2];
	log_buf = (const char*) desc[3];

	log_buf_len = ( desc[4] & ~( 1 << 31 ) );
	if (log_buf_len == 0)
	     return;
	if ( log_buf_len == 0 || ( log_buf_len & (log_buf_len - 1) ) != 0 ) {
		fprintf(stderr, "WARN: klog buffer length not a "
			"power of two, ignoring\n");
		return;
	}

	if ( logged_chars > log_buf_len ) {
		fprintf(stderr, "WARN: inconsistent klog data, ignoring\n");
		return;
	}
	if (logged_chars == 0) {
		DPRINTF("empty klog buffer, ignoring\n");
		return; /* empty log buffer */
	}

	ignore = ( desc[4] >> 31 );

	DPRINTF("klog data: %lu %lu %p %lu %i\n",
		log_end, logged_chars, log_buf, log_buf_len, ignore);

	if ( ignore ) {
		DPRINTF("klog buffer marked for ignoring\n");
		return;
	}

	save_log(log_end, logged_chars, log_buf, log_buf_len);
}

void klog_dump_reset(void)
{
	volatile ulong *desc =
		(volatile ulong *)( KLOG_DUMP_ADDR );
	int i;

	DPRINTF("resetting klog dump data\n");

	desc[0] = KLOG_MAGIC;
	for (i = 1; i < 5; i++)
	     desc[i] = 0;
}

#endif /* CONFIG_KLOG_DUMP */
